import datetime
import os
import time
import traceback
from urllib.error import HTTPError

from slack_sdk.webhook import WebhookClient

from calendar_event_engine.db.db_cache import SQLiteDB
from calendar_event_engine.filter import normalize_generic_event
from calendar_event_engine.logger import create_logger_from_designated_logger
from calendar_event_engine.parser.submission import get_runner_submission
from calendar_event_engine.publishers.abc_publisher import Publisher
from calendar_event_engine.scrapers.abc_scraper import Scraper
from calendar_event_engine.types.custom_scraper import CustomScraperJob
from calendar_event_engine.types.submission import GroupPackage, AllEventsFromAGroup, GroupEventsKernel
from calendar_event_engine.types.submission_handlers import RunnerSubmission
from calendar_event_engine.scrapers.google_calendar.api import ExpiredToken

logger = create_logger_from_designated_logger(__name__)


def _runner(runner_submission: RunnerSubmission, custom_scrapers: dict[Publisher, list[CustomScraperJob]] = None):
    continue_scraping = True
    num_retries = 0
    theres_an_expired_token = False
    while continue_scraping and num_retries < 5:
        try:
            submitted_publishers: {Publisher: list[GroupPackage]} = runner_submission.publishers
            for publisher in submitted_publishers.keys():
                publisher: Publisher
                publisher.connect()
                group_package: GroupPackage
                for group_package in submitted_publishers[publisher]:
                    logger.info(f"Reading Group Package: {group_package.package_name}")

                    for scraper_type in group_package.scraper_type_and_kernels.keys():
                        scraper: Scraper = runner_submission.respective_scrapers[scraper_type]
                        try:
                            scraper.connect_to_source()
                            group_event_kernels: list[GroupEventsKernel] = group_package.scraper_type_and_kernels[
                                scraper_type]
                            for event_kernel in group_event_kernels:
                                event_kernel: GroupEventsKernel
                                try:
                                    events: list[AllEventsFromAGroup] = scraper.retrieve_from_source(event_kernel)
                                except HTTPError as err:
                                    if err.code == 404:
                                        logger.warning(f"The following group is no longer available: {event_kernel.group_name}")
                                    else:
                                        raise
                                normalize_generic_event(events)
                                publisher.upload(events)
                            scraper.close_connection_to_source()
                        except ExpiredToken:
                            theres_an_expired_token = True
                            logger.warning("Expired token.json needs to be replaced. Will continue scraping other types")
                            continue
                publisher.close()


            if custom_scrapers:
                for publisher in custom_scrapers.keys():
                    publisher.connect()
                    for scrap in custom_scrapers[publisher]:
                        try:
                            logger.info(f"Using custom scraper {scrap.scraper_name}")
                            scrap.custom_scraper.connect_to_source()
                            events = scrap.custom_scraper.retrieve_from_source(None)
                            normalize_generic_event(events)
                            publisher.upload(events)
                        except Exception as custom_err:
                            logger.error("Exception for custom scraper: " + scrap.scraper_name, custom_err)
                    publisher.close()
            continue_scraping = False
        except HTTPError as err:
            if err.code == 500 and err.reason.lower() == 'Too many requests'.lower():
                num_retries += 1
                logger.warning("Going to sleep then retrying to scrape. Retry Num: " + num_retries)
                time.sleep(120)
    if theres_an_expired_token:
        raise ExpiredToken

def _days_to_sleep(days):
    now = datetime.datetime.now()
    seconds_from_zero = (now.hour * 60 * 60)
    time_at_2am = 2 * 60 * 60
    time_to_sleep = time_at_2am
    if seconds_from_zero > time_at_2am:
        time_to_sleep = ((23 * 60 * 60) - seconds_from_zero) + time_at_2am
    else:
        time_to_sleep = time_at_2am - seconds_from_zero
    
    return time_to_sleep + (60 * 60 * 24 * days)


def _produce_slack_message(color, title, text, priority):
    return {
            "color": color,
            "author_name": "CTEvent Scraper",
            "author_icon": "https://ctgrassroots.org/favicon.ico",
            "title": title,
            "title_link": "google.com",
            "text": text,
            "fields": [
                {
                    "title": "Priority",
                    "value": priority,
                    "short": "false"
                }
            ],
            "footer": "CTEvent Scraper",
        }

def start_event_engine(remote_json_url: str, cache_db: SQLiteDB, slack_webhook: WebhookClient = None, custom_scrapers: dict[Publisher, list[CustomScraperJob]] = None,
                       test_mode: bool = False):
    logger.info("Scraper Started")
    sleeping = 2
    while True:
        #####################
        # Create Submission #
        #####################

        submission: RunnerSubmission = get_runner_submission(remote_json_url, test_mode, cache_db)

        ######################
        # Execute Submission #
        ######################
        time_to_sleep = _days_to_sleep(sleeping)
        logger.info("Scraping")
        try:
            _runner(submission, custom_scrapers)
            logger.info("Sleeping " + str(sleeping) + " Days Until Next Scrape")
        except ExpiredToken:
            time_to_sleep = _days_to_sleep(2)
            logger.info("Sleeping " + str(sleeping) + " Days Until Next Scrape")
            if slack_webhook is not None:
                response = slack_webhook.send(attachments=[
                    _produce_slack_message("#e6e209", "Expired Token", "Replace token.json", "Medium")
                ])

        except Exception as e:
            logger.error("Unknown Error")
            logger.error(e)
            logger.error(traceback.format_exc())
            logger.error("Going to Sleep for 7 days")
            if slack_webhook is not None:
                slack_webhook.send(attachments=[
                    _produce_slack_message("#ab1a13", "Event Scraper Unknown Error", "Check logs for error.", "High")
                ])
            time_to_sleep = _days_to_sleep(7)

        cache_db.close()
        time.sleep(time_to_sleep)


if __name__ == "__main__":
    env_webhook = None if os.environ.get("SLACK_WEBHOOK") is None else WebhookClient(os.environ.get("SLACK_WEBHOOK"))
    env_test_mode = False if "TEST_MODE" not in os.environ else True
    submission_json_path = os.getenv("RUNNER_SUBMISSION_JSON_PATH")
    if submission_json_path is None:
        err = NotImplementedError("Did not find JSON submission path")
        logger.error(err)
        raise err


    start_event_engine(submission_json_path, SQLiteDB(env_test_mode), env_webhook, None, env_test_mode)





import os


from calendar_event_engine.db.db_cache import SQLiteDB
from calendar_event_engine.logger import create_logger_from_designated_logger
from calendar_event_engine.parser.package import get_group_package
from calendar_event_engine.publishers.mobilizon.api import logger
from calendar_event_engine.scrapers.abc_scraper import Scraper
from calendar_event_engine.scrapers.google_calendar.api import GCalAPI
from calendar_event_engine.types.generics import GenericEvent
from calendar_event_engine.types.submission import GroupEventsKernel, AllEventsFromAGroup, ScraperTypes

logger = create_logger_from_designated_logger(__name__)

class GoogleCalendarScraper(Scraper):
    def close_connection_to_source(self) -> None:
        self.google_calendar_api.close()

    def get_source_type(self):
        return ScraperTypes.GOOGLE_CAL

    google_calendar_api: GCalAPI
    cache_db: SQLiteDB
    def __init__(self, cache_db: SQLiteDB):
        self.cache_db = cache_db
        self.google_calendar_api = GCalAPI()

    def _get_specific_calendar_events(self, google_calendar_id, group_kernel: GroupEventsKernel):
        last_uploaded_event_date = None
        if not self.cache_db.no_entries_with_source_id(google_calendar_id):
            last_uploaded_event_date = self.cache_db.get_last_event_date_for_source_id(google_calendar_id)

        events: list[GenericEvent] = self.google_calendar_api.getAllEventsAWeekFromNow(
            calendarId=google_calendar_id, eventKernel=group_kernel.event_template,
            checkCacheFunction=self.cache_db.entry_already_in_cache,
            dateOfLastEventScraped=last_uploaded_event_date)

        return events


    def retrieve_from_source(self, group_event_kernel: GroupEventsKernel) -> list[AllEventsFromAGroup]:


        all_events: list[AllEventsFromAGroup] = []
        logger.info(f"Getting events from calendar {group_event_kernel.group_name}")
        for google_calendar_id in group_event_kernel.calendar_ids:
            events = self._get_specific_calendar_events(google_calendar_id, group_event_kernel)
            all_events.append(AllEventsFromAGroup(events, group_event_kernel, google_calendar_id))

        return all_events


    ############################
    # Used Mostly for Testing ##
    ############################
    def get_gcal_events_for_specific_group_and_upload_them(self, calendar_group: str):
        google_calendars: list[GroupEventsKernel] = get_group_package(f"{os.getcwd()}/src/scrapers/GCal.json")
        logger.info(f"\nGetting events from calendar {calendar_group}")
        gCal: GroupEventsKernel
        all_events: list[GenericEvent] = []
        for gCal in google_calendars:
            if gCal.group_name == calendar_group:
                for googleCalendarID in gCal.calendar_ids:
                    all_events += self._get_specific_calendar_events(googleCalendarID, gCal)
        return all_events

    def connect_to_source(self):
        use_oidc = os.environ.get("USE_OIDC_TOKEN")
        if use_oidc:
            token_path = f"{os.getcwd()}/config/token.json" if "GOOGLE_API_TOKEN_PATH" not in os.environ else os.environ.get("GOOGLE_API_TOKEN_PATH")
            self.google_calendar_api.init_calendar_read_client_browser(token_path)
        else:
            self.google_calendar_api.init_calendar_read_client_adc()


import copy
import os
import urllib.request
from datetime import datetime, timedelta, timezone, date

import icalendar
import validators
from icalendar.cal import Calendar

from calendar_event_engine.db.db_cache import SQLiteDB
from calendar_event_engine.logger import create_logger_from_designated_logger
from calendar_event_engine.publishers.mobilizon.api import logger
from calendar_event_engine.scrapers.abc_scraper import Scraper
from calendar_event_engine.types.generics import GenericEvent, GenericAddress
from calendar_event_engine.types.submission import ScraperTypes, AllEventsFromAGroup, GroupEventsKernel
from calendar_event_engine.utils.location import find_geolocation_from_address

logger = create_logger_from_designated_logger(__name__)


class ICALScraper(Scraper):
    def close_connection_to_source(self) -> None:
        pass

    def get_source_type(self):
        return ScraperTypes.GOOGLE_CAL

    cache_db: SQLiteDB
    def __init__(self, cache_db: SQLiteDB):
        self.cache_db = cache_db

    def retrieve_from_source(self, group_event_kernel: GroupEventsKernel) -> list[AllEventsFromAGroup]:
        all_events: list[AllEventsFromAGroup] = []
        logger.info(f"Getting events from calendar {group_event_kernel.group_name}")
        for ical_id in group_event_kernel.calendar_ids:
            with urllib.request.urlopen(ical_id) as f:
                calendar = icalendar.Calendar.from_ical(f.read())
                events = _hydrate_event_template(calendar, group_event_kernel.event_template)
                all_events.append(AllEventsFromAGroup(events, group_event_kernel, ical_id))

        return all_events

    # Get only events 1 week from today, and that are confirmed

    def connect_to_source(self):
        pass


def _hydrate_event_template(calendar: Calendar, event_kernel: GenericEvent) -> list[GenericEvent]:
    week_from_now = datetime.now(timezone.utc) + timedelta(days=7)
    events = []
    for event in calendar.walk('VEVENT'):
        event_template = copy.deepcopy(event_kernel)
        start = event.get("DTSTART").dt
        if type(start) == date:
            start = datetime.combine(start, datetime.min.time(), timezone.utc)
        end = event.get("DTEND").dt
        if type(end) == date:
            end = datetime.combine(end, datetime.min.time(), timezone.utc)
        summary = str(event.get("SUMMARY"))
        status = event.get("STATUS")
        over_a_week = week_from_now < start
        before_today = start < datetime.now(timezone.utc)

        # Have to search entire list because events aren't organized
        if (over_a_week or before_today) and os.getenv("TEST") != "True":
            continue
        elif status == "CONFIRMED":
            event_template.title = summary
            event_template.begins_on = start.isoformat()
            event_template.ends_on = end.isoformat()
            if 'ATTACH' in event and validators.url(event.get("ATTACH")):
                event_template.picture = event.get("ATTACH")
            grabbed_description = "" if "DESCRIPTION" not in event else str(event.get("DESCRIPTION"))
            notif = ""
            event_template.online_address = event_template.online_address if "URL" not in event else str(event.get("URL"))
            if "LOCATION" in event:
                parsed_location = _parse_retrieved_location(str(event.get("LOCATION")), event_template.physical_address)
                event_template.physical_address, notif = find_geolocation_from_address(parsed_location,
                                                                               event_template.physical_address,
                                                                               event_template.title)
            event_template.description = f"Automatically scraped by Event Bot {notif}: \n\n{grabbed_description}"
            events.append(event_template)

    return events


def _parse_retrieved_location(location: str, default_location: GenericAddress) -> GenericAddress:
    if location is None:
        logger.debug("No location provided, using default")
        return default_location
    tokens = location.split(",")
    address: GenericAddress
    match len(tokens):
        case 1:
            return default_location
        case 2:
            return default_location
        case 3:
            address = GenericAddress(locality=tokens[2], street=tokens[1])
        case 4:
            address = GenericAddress(locality=tokens[2], street=tokens[1],
                                              region=tokens[3])
        case 5:
            address = GenericAddress(locality=tokens[2], postalCode=tokens[4], street=tokens[1],
                                            region=tokens[3])
        case _:
            return default_location
    return address





import copy
from datetime import datetime, timedelta


from calendar_event_engine.logger import create_logger_from_designated_logger
from calendar_event_engine.scrapers.abc_scraper import Scraper
from calendar_event_engine.types.generics import GenericEvent
from calendar_event_engine.types.submission import ScraperTypes, GroupEventsKernel, AllEventsFromAGroup

logger = create_logger_from_designated_logger(__name__)


class StaticScraper(Scraper):

    def close_connection_to_source(self) -> None:
        pass

    def get_source_type(self):
        return ScraperTypes.STATIC

    def connect_to_source(self):
        pass

    def retrieve_from_source(self, group_kernel: GroupEventsKernel) -> [AllEventsFromAGroup]:
        logger.info(f"Getting static events: {group_kernel.group_name}")
        events: [GenericEvent] = hydrate_event_template_with_legitimate_times(group_kernel)
        event: GenericEvent
        for event in events:
            event.description = f"Automatically scraped by event bot: \n\n{event.description} \n\n Source for farmer market info: https://portal.ct.gov/doag/adarc/adarc/farmers-market-nutrition-program/authorized-redemption-locations"
        return [AllEventsFromAGroup(events, group_kernel, group_kernel.group_name)]



def hydrate_event_template_with_legitimate_times(group_kernel: GroupEventsKernel) -> [GenericEvent]:
    """
    Updating the initial default times from the static event to their relevant times for the week, unless
    the end date has been reached.
    """

    #############################################################
    # There can multiple days in a week when an event can occur #
    #############################################################
    times = group_kernel.default_time_info.default_times

    generated_events = []

    end_date = datetime.fromisoformat(group_kernel.default_time_info.end_time)
    now = datetime.utcnow().astimezone()

    if now.date() <= end_date.date():
        for t in times:
            event: GenericEvent = copy.deepcopy(group_kernel.event_template)
            start_time = datetime.fromisoformat(t[0])
            end_time = datetime.fromisoformat(t[1])

            time_difference_weeks = (now - start_time).days // 7  # Floor division that can result in week prior event

            start_time += timedelta(weeks=time_difference_weeks)
            end_time += timedelta(weeks=time_difference_weeks)

            if start_time < now:
                start_time += timedelta(weeks=1)
                end_time += timedelta(weeks=1)
                if start_time > end_date:
                    return []

            event.begins_on = start_time.astimezone().isoformat()
            event.ends_on = end_time.astimezone().isoformat()

            generated_events.append(event)

        return generated_events

    logger.info(f"Static Event {group_kernel.group_name} Has Expired")
    return []



import json
import urllib

from calendar_event_engine.types.generics import GenericAddress, GenericEvent
from calendar_event_engine.types.submission import ScraperTypes, GroupPackage, TimeInfo, GroupEventsKernel


def none_if_not_present(x, dictionary):
    return None if x not in dictionary else dictionary[x]


def retrieve_source_type(source_type_string) -> ScraperTypes:
    match source_type_string:
        case "STATIC":
            return ScraperTypes.STATIC
        case "GOOGLE_CAL":
            return ScraperTypes.GOOGLE_CAL
        case "ICAL":
            return ScraperTypes.ICAL
        case "CUSTOM":
            return ScraperTypes.CUSTOM
    raise TypeError("Expected a string that resolves to a scraper type, instead got: " + source_type_string)


def get_group_package(json_path: str) -> GroupPackage:
    group_schema: dict = None
    with urllib.request.urlopen(json_path) as f:
        group_schema = json.load(f)

    group_package: GroupPackage = GroupPackage({}, none_if_not_present("name", group_schema),
                                               none_if_not_present("description", group_schema))

    for group_name, group_info in group_schema["groupKernels"].items():

        event_address = None if "defaultLocation" not in group_info else GenericAddress(**group_info["defaultLocation"])
        event_kernel = GenericEvent(group_info["publisherInfo"], none_if_not_present("title", group_info),
                                    none_if_not_present("beginsOn", group_info),
                                    none_if_not_present("defaultDescription", group_info),
                                    none_if_not_present("endsOn", group_info),
                                    group_info["onlineAddress"], none_if_not_present("phoneAddress", group_info),
                                    event_address)

        calendar_ids = group_info["calendarIDs"]
        scraper_type: ScraperTypes = retrieve_source_type(group_info["calendarType"])
        time_info = None
        if scraper_type == ScraperTypes.STATIC:
            time_info = TimeInfo(group_info["defaultTimes"], group_info["endDate"])



        if scraper_type not in group_package.scraper_type_and_kernels:
            group_package.scraper_type_and_kernels[scraper_type] = []

        group_package.scraper_type_and_kernels[scraper_type].append(
            GroupEventsKernel(event_kernel, group_name, calendar_ids=calendar_ids,
                              scraper_type=scraper_type, json_source_url=json_path, time_info=time_info))

    return group_package



from enum import Enum

from calendar_event_engine.types.generics import GenericEvent


class ScraperTypes(Enum):
    STATIC = "JSON"
    GOOGLE_CAL = "Google Calendar"
    ICAL = "ICAL"
    CUSTOM = "CUSTOM"


class PublisherTypes(Enum):
    MOBILIZON = "Mobilizon"


class TimeInfo:
    """
    Used to hydrate the static event schema,
    """
    default_times: []
    end_time: str

    def __init__(self, default_times, end_time):
        self.default_times = default_times
        self.end_time = end_time


class GroupEventsKernel:
    """
    All the information required to scrape a groups' information.
    An individual group can have multiple calendars.
    """
    event_template: GenericEvent
    group_name: str
    calendar_ids: list[str]
    scraper_type: ScraperTypes
    json_source_url: str
    default_time_info: TimeInfo

    def __init__(self, event, group_name, calendar_ids, scraper_type, json_source_url,
                 time_info = None):
        self.event_template = event
        self.calendar_ids = calendar_ids
        self.group_name = group_name
        self.scraper_type = scraper_type
        self.json_source_url = json_source_url
        self.default_time_info = time_info

    def __eq__(self, other):
        if not isinstance(other, GroupEventsKernel):
            return False
        compare = self.group_name == other.group_name and self.event_template == other.event_template
        compare = compare and self.calendar_ids == other.calendar_ids and self.scraper_type == other.scraper_type
        compare = compare and self.json_source_url == other.json_source_url
        return compare


class GroupPackage:
    """
    A bundle of scrapers (ICal, Google, etc...) and the groups (ex. Salsa Club, Nature Volunteers, Car Enthusiasts, etc...)
    that information will be retrieved from.
    """
    package_name: str
    description: str

    scraper_type_and_kernels: dict[ScraperTypes, list[GroupEventsKernel]]

    def __init__(self, scrapers_with_group_kernels, package_name, description):
        self.scraper_type_and_kernels = scrapers_with_group_kernels
        self.package_name = package_name
        self.description = description


class AllEventsFromAGroup:
    """
    Wrapper for the events scraped for a particular calendar id.
    Args:
        events: All the events scrapped.
        event_kernel: All information about the group which being scrapped.
        source_id: The specific calendar where all the events scrapped originated from.
    """
    events: list[GenericEvent] = None
    eventKernel: GroupEventsKernel = None
    calendar_id: str = ""

    def __init__(self, events: list[GenericEvent], event_kernel: GroupEventsKernel, source_id: str):
        self.events = events
        self.eventKernel = event_kernel
        self.calendar_id = source_id



class GenericAddress:
    """
    Object used to store address information.
    Args:
        geom: Geo-coordinates for the point where the address is
        region: The state of this address
        locality: Town
        postalCode: ZipCode
        street: Street
        country: Country
    """
    geom: str
    locality: str
    postalCode: str
    street: str
    country: str
    region: str
    timezone: str
    description: str

    def __init__(self, geom: str = None, locality: str = None, postalCode: str = None, street:str=None, country: str="United States",
                 region: str = None, timezone:str = "America/New_York", description = ""):
        self.geom = geom
        self.locality = locality
        self.postalCode = postalCode
        self.street = street
        self.country = country
        self.region = region
        self.timezone = timezone
        self.description = description

    def __eq__(self, other):
        if type(other) != GenericAddress:
            return False
        other: GenericAddress
        local_postal_street: bool = other.locality == self.locality and other.postalCode == self.postalCode and other.street == self.street
        country_region_timezone: bool = other.country == self.country and other.region == self.region and other.timezone == self.timezone
        return local_postal_street and country_region_timezone and other.geom == self.geom and other.description == self.description

    def __str__(self):
        return f"{self.geom}, {self.locality}, {self.postalCode}, {self.street}, {self.region}"

class GenericEvent:
    """
    Object used to store event related information.
    Time is parsed with the local time zone available. Datetime + timezone
    """
    publisher_specific_info: dict

    title: str
    description: str
    begins_on: str
    ends_on: str
    picture: str
    online_address: str
    phone_address: str
    physical_address: GenericAddress

    def __init__(self, publisher_specific_info: dict, title: str, begins_on: str, description: str = None,
                 ends_on: str = None, online_address: str = None,
                 phone_address: str = None, physical_address: GenericAddress = None, picture: str = None):
        self.publisher_specific_info = publisher_specific_info
        self.title = title
        self.description = description
        self.begins_on = begins_on
        self.ends_on = ends_on
        self.online_address = online_address
        self.phone_address = phone_address
        self.physical_address = physical_address
        self.picture = picture

    @classmethod
    def default(cls):
        return cls(None, None, None)

    def __eq__(self, other):
        if type(other) != GenericEvent:
            return False
        else:
            other: GenericEvent
            time_and_address: bool = other.ends_on == self.ends_on and other.begins_on == self.begins_on and other.online_address == self.online_address and other.phone_address == self.phone_address and other.physical_address == self.physical_address
            description_and_title: bool = other.title == self.title and other.description == self.description
            return time_and_address and description_and_title and other.publisher_specific_info == self.publisher_specific_info

    def __str__(self):
        return f"Title: {self.title}, Begins On: {self.begins_on}, Ends On: {self.ends_on}\nDescription: {self.description}\nAddress: {self.physical_address}\nPicture URL: {self.picture}"





import sqlite3


class EventSource:
    uuid: str
    websiteURL: str
    calendar_id: str
    sourceType: str

    def __init__(self, uuid:str, website_url: str, calendar_id:str, source_type: str):
        """
        Args:
            uuid(str): The publishers UUID for the event uploaded
            website_url(str): The events website of origin
            calendar_id(str): The calendar where the event was sourced from
            source_type(str): Source type
        """
        self.uuid = uuid
        self.websiteURL = website_url
        self.calendar_id = calendar_id
        self.sourceType = source_type


class EventSourceDriver:
    event_source_table_name = "event_source"

    def __init__(self, db_connection: sqlite3.Connection):
        db_cursor = db_connection.cursor()
        db_cursor.execute(f"""CREATE  TABLE IF NOT EXISTS {self.event_source_table_name}
                                  (id, uuid text, websiteURL text, calendar_id text, sourceType text, 
                                  FOREIGN KEY (id) REFERENCES uploaded_events(id) ON DELETE CASCADE)""")
        db_connection.commit()

    def select_all_from_event_source_table(self, db_connection: sqlite3.Connection) -> sqlite3.Cursor:
        db_cursor = db_connection.cursor()
        res = db_cursor.execute(f"SELECT * FROM {self.event_source_table_name}")
        return res

    def insert_uploaded_event(self, db_connection: sqlite3.Connection, foreign_key: int, event_source: EventSource):
        db_cursor: sqlite3.Cursor = db_connection.cursor()
        event_source_row = (foreign_key, event_source.uuid, event_source.websiteURL, event_source.calendar_id, event_source.sourceType)

        db_cursor.execute(f"INSERT INTO {self.event_source_table_name} VALUES (?, ?, ? , ?, ?)", event_source_row)
        db_connection.commit()






from calendar_event_engine.db.db_cache import SQLiteDB
from calendar_event_engine.publishers.abc_publisher import Publisher
from calendar_event_engine.scrapers.abc_scraper import Scraper
from calendar_event_engine.types.submission import ScraperTypes, GroupPackage


class RunnerSubmission:
    respective_scrapers: dict[ScraperTypes, Scraper]
    publishers: dict[Publisher, list[GroupPackage]]
    def __init__(self, submitted_db: SQLiteDB,
                 submitted_publishers: dict[Publisher, list[GroupPackage]],
                 test: bool,
                 respective_scrapers: dict[ScraperTypes, Scraper]):
        self.cache_db = submitted_db
        self.test = test
        self.publishers = submitted_publishers
        self.respective_scrapers = respective_scrapers






import copy
import os
from datetime import datetime, timedelta

import google.auth
from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials
from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.discovery import build, Resource
from googleapiclient.errors import HttpError

from calendar_event_engine.logger import create_logger_from_designated_logger
from calendar_event_engine.types.generics import GenericAddress, GenericEvent
from calendar_event_engine.utils.location import find_geolocation_from_address

logger = create_logger_from_designated_logger(__name__)

# Subscribe to the calendars
# https://webapps.stackexchange.com/questions/5217/how-can-i-find-the-subscribe-url-from-the-google-calendar-embed-source-code
# http://www.google.com/calendar/feeds/HERE/public/basic

# Bike Shop Source 
# bsbc.co_c4dt5esnmutedv7p3nu01aerhk@group.calendar.google.com

# Save The Sound Source
# ctenvironment@gmail.com

# Google API Documentation
# https://console.cloud.google.com/apis/credentials/consent
# https://developers.google.com/calendar/api/quickstart/python
# https://developers.google.com/resources/api-libraries/documentation/calendar/v3/python/latest/calendar_v3.events.html#list


class ExpiredToken(Exception):
    pass


class GCalAPI:
    _apiClient: Resource
    
    def __init__(self):
        pass
    
    def init_calendar_read_client_browser(self, token_path: str):
        SCOPES = ["https://www.googleapis.com/auth/calendar.readonly"]
        credentialTokens = None
        if os.path.exists(token_path):
            credentialTokens = Credentials.from_authorized_user_file(token_path, SCOPES)
        
        if not credentialTokens or not credentialTokens.valid:
            try: 
                if credentialTokens and credentialTokens.expired and credentialTokens.refresh_token:
                    credentialTokens.refresh(Request())
                else:
                    flow = InstalledAppFlow.from_client_secrets_file(
                            f"{os.getcwd()}/config/OAuthClientApp.json", SCOPES
                        )
                    credentialTokens = flow.run_local_server(port=9000)
                
                # When refreshed authentication token needs to be re-written, and if authenticating for the first time it needs to be just written
                with open(token_path, "w") as tokenFile:
                    tokenFile.write(credentialTokens.to_json())
            except Exception:
                raise ExpiredToken
        self._apiClient = build("calendar", "v3", credentials=credentialTokens)
        logger.info("Logged in Google Cal Browser")

    
    def init_calendar_read_client_adc(self):
        logger.info("Logged In Google Cal ADC")
        credentials, projectID = google.auth.default()
        
        self._apiClient = build("calendar", "v3", credentials=credentials)
    

    def getAllEventsAWeekFromNow(self, eventKernel: GenericEvent, calendarId: str,
                                 checkCacheFunction,
                                 dateOfLastEventScraped: datetime = None) -> list[GenericEvent]:
        """Get events all events for that specific calender a week from today.

        Args:
            service (Resource): Calender resource available from creating a client with 'read calender' scope
            calendarId (str): UUID of calender. Same ID used to subscribe to a calender
            dateOfLastEventScraped (str, optional): UTC date with timedelta following isoformat. Have to add 'Z' to isoformat to indicate its UTC time. Defaults to None.
        """
        try:
            # Call the Calendar API, 'Z' indicates UTC time
            stringDateLastEvent = datetime.utcnow().astimezone().isoformat()
            if dateOfLastEventScraped is not None: 
                stringDateLastEvent = dateOfLastEventScraped.isoformat()
            logger.debug(f"Time of last event: {stringDateLastEvent}")
            weekFromNow = datetime.utcnow().astimezone() + timedelta(days=7)
            weekFromNow = weekFromNow.isoformat()
            
            events_result = (
                self._apiClient.events()
                .list(
                    calendarId=calendarId,
                    timeMin=stringDateLastEvent,
                    timeMax=weekFromNow,
                    singleEvents=True,
                    orderBy="startTime",
                )
                .execute()
            )
            googleEvents = events_result.get("items", [])
            if len(googleEvents) == 0:
                logger.info(f"No upcoming events for calendarID {calendarId}\n")
                return googleEvents

            events = []
            for googleEvent in googleEvents:
                _process_google_event(googleEvent=googleEvent, eventsToUpload=events,
                                    checkCacheForEvent=checkCacheFunction,calendarId=calendarId,
                                    eventKernel=copy.deepcopy(eventKernel))
            
            return events
        except HttpError as error:
            logger.error(f"An error occurred: {error}")
    
    def close(self):
        self._apiClient.close()





def _process_google_event(googleEvent: dict, eventsToUpload: [], checkCacheForEvent, 
                          calendarId: str, eventKernel: GenericEvent):
    
    starTimeGoogleEvent = googleEvent["start"].get("dateTime")
    endTimeGooglEvent = googleEvent["end"].get("dateTime")
    title = googleEvent.get("summary")
    description = googleEvent.get("description")

    
    if None not in [starTimeGoogleEvent, endTimeGooglEvent, title, description]:
        startDateTime = datetime.fromisoformat(starTimeGoogleEvent.replace('Z', '+00:00')).astimezone()
        endDateTime = datetime.fromisoformat(endTimeGooglEvent.replace('Z', '+00:00')).astimezone()
        if not checkCacheForEvent(startDateTime.isoformat(), title, calendarId):
            eventAddress, default_location_notif = find_geolocation_from_address(parse_google_location(googleEvent.get("location"), eventKernel.physical_address), eventKernel.physical_address, title)
            eventKernel.begins_on = startDateTime.isoformat()
            eventKernel.ends_on = endDateTime.isoformat()
            eventKernel.physical_address = eventAddress
            eventKernel.title = title
            eventKernel.description = f"Automatically scraped by Event Bot {default_location_notif}: \n\n{description}"
            eventsToUpload.append(eventKernel)
            

def parse_google_location(location:str, default_location: GenericAddress):
    if location is None:
        logger.debug("No location provided, using default")
        return default_location
    tokens = location.split(",")
    address: GenericAddress
    match len(tokens):
        case 1:
            return default_location
        case 2:
            return default_location
        case 3:
            address = GenericAddress(locality=tokens[0], postalCode=tokens[1], street="", country=tokens[2])
        case 4:
            address = GenericAddress(locality=tokens[1], postalCode=tokens[2], street=tokens[0], country=tokens[3])
        case 5:
            address = GenericAddress(locality=tokens[2], postalCode=tokens[3], street=tokens[1], country=tokens[4])
        case _:
            return default_location

    return address

    

if __name__ == "__main__":
    gcal = GCalAPI()
    google_token_path = os.environ.get("GOOGLE_API_TOKEN_PATH")
    gcal.init_calendar_read_client_browser(google_token_path)




from abc import ABC, abstractmethod

from calendar_event_engine.types.submission import GroupEventsKernel, AllEventsFromAGroup, ScraperTypes


class Scraper(ABC):
    """
    The class used to retrieve information from a particular source type.
    """
    @abstractmethod
    def connect_to_source(self) -> None:
        pass

    @abstractmethod
    def retrieve_from_source(self, event_kernel: GroupEventsKernel=None) -> list[AllEventsFromAGroup]:
        """
        Takes GroupEventKernel and returns list[EventsToUploadFromCalendarID]
        For custom scrapers, ignore any input and directly scrape from the resource to create EventsToUploadFromCalendarID.
        """
        pass

    @abstractmethod
    def close_connection_to_source(self) -> None:
        pass

    @abstractmethod
    def get_source_type(self) -> ScraperTypes:
        pass





from calendar_event_engine.publishers.abc_publisher import Publisher
from calendar_event_engine.scrapers.abc_scraper import Scraper


class CustomScraperJob:
    """
    A scraper that implements the 'Scraper' abstract base class, and will
    be executed by the event scraper engine
    """
    scraper_name: str
    description: str
    custom_scraper: Scraper

    def __init__(self, scraper_name: str, description: str, custom_scraper: Scraper):
        self.custom_scraper = custom_scraper
        self.description = description
        self.scraper_name = scraper_name


import os
import sqlite3
from datetime import datetime

from calendar_event_engine.db.combine_driver import CombineDBDriver
from calendar_event_engine.db.event_source_driver import EventSourceDriver, EventSource
from calendar_event_engine.db.uploaded_events_driver import UploadedEventsDriver, UploadedEventRow
from calendar_event_engine.logger import create_logger_from_designated_logger

from sqlite3 import Cursor

logger = create_logger_from_designated_logger(__name__)


class SQLiteDB:
    sql_db_connection: sqlite3.Connection
    es_driver: EventSourceDriver
    ue_driver: UploadedEventsDriver
    com_driver: CombineDBDriver
    __fake_UUID: int = 0

    def __init__(self, test_mode: bool = False):
        if test_mode:
            self.sql_db_connection = sqlite3.connect(":memory:")
        else:
            cache_db_path = os.environ.get("CACHE_DB_PATH")
            if cache_db_path is not None:
                self.sql_db_connection = sqlite3.connect(cache_db_path + "/event_cache.db")
            else:
                self.sql_db_connection = sqlite3.connect("../event_cache.db")

        self.ue_driver = UploadedEventsDriver(self.sql_db_connection)
        self.es_driver = EventSourceDriver(self.sql_db_connection)
        self.com_driver = CombineDBDriver()

    def close(self):
        self.sql_db_connection.close()

    def delete_all_month_old_events(self):
        self.ue_driver.delete_all_month_old_events(self.sql_db_connection)

    def select_all_rows_with_calendar_id(self, source) -> Cursor:
        return self.com_driver.select_all_rows_with_calendar_id(self.sql_db_connection, source)
    
    def get_last_event_date_for_source_id(self, calendar_id) -> datetime:
        return self.com_driver.get_last_event_date_for_source_id(self.sql_db_connection, calendar_id)
    
    def no_entries_with_source_id(self, calendar_id: str) -> bool:
        res = self.select_all_rows_with_calendar_id(calendar_id)
        return len(res.fetchall()) == 0
    
    def entry_already_in_cache(self, date:str, title:str, source_id:str) -> bool:
        return self.com_driver.entry_already_in_cache(self.sql_db_connection, date, title, source_id)

    def insert_uploaded_event(self, row_to_add: UploadedEventRow, event_source: EventSource):
        primary_key = self.ue_driver.insert_uploaded_event(self.sql_db_connection, row_to_add)
        self.es_driver.insert_uploaded_event(self.sql_db_connection, primary_key, event_source)
    
    def select_all_from_upload_table(self) -> sqlite3.Cursor:
        return self.ue_driver.select_all_from_upload_table(self.sql_db_connection)

    def select_all_from_event_source_table(self) -> sqlite3.Cursor:
        return self.es_driver.select_all_from_event_source_table(self.sql_db_connection)
    
        
        
    


import sqlite3

class UploadedEventRow:
    uuid: str
    user_id: str
    title: str
    date: str
    groupID: int
    groupName: str

    def __init__(self, uuid: str, user_id: str, title: str, date: str, group_id: int, group_name: str):
        """_summary_

        Args:
            uuid (str): The publishers UUID for the event uploaded
            user_id (str): The publishers id of the user who uploaded this event
            title (str): Title of the event
            date (str): Has to be of format ISO8601 that is YYYY-MM-DD HH:MM:SS.SSS. Can also have T in the center if desired.
            group_id (int): The publishers UUID for the group affiliated with this event
            group_name (str): Name of the group affiliated with this event
        """

        self.uuid = uuid
        self.user_id = user_id
        self.title = title
        self.date = date
        self.groupID = group_id
        self.groupName = group_name

class UploadedEventsDriver:
    uploaded_events_table_name = "uploaded_events"


    def __init__(self, db_connection: sqlite3.Connection):
        db_cursor = db_connection.cursor()
        db_cursor.execute(f"""CREATE TABLE IF NOT EXISTS {self.uploaded_events_table_name} 
                                  (id INTEGER PRIMARY KEY, uuid text, user_id text, 
                                  title text, date DATE, group_id integer, group_name text)""")

        # https://www.sqlite.org/lang_datefunc.html
        # Uses built in date time function
    def delete_all_month_old_events(self, db_connection: sqlite3.Connection):
        db_cursor = db_connection.cursor()
        db_cursor.execute(
            f"DELETE FROM {self.uploaded_events_table_name} WHERE datetime(date) < datetime('now', '-1 month')")
        db_connection.commit()

    def select_all_from_upload_table(self, db_connection: sqlite3.Connection) -> sqlite3.Cursor:
        db_cursor = db_connection.cursor()
        res = db_cursor.execute(f"SELECT * FROM {self.uploaded_events_table_name}")
        return res

    def insert_uploaded_event(self, db_connection: sqlite3.Connection, row_to_add: UploadedEventRow) -> int:
        """
        Returns the primary key for the row when it gets inserted.
        """
        db_cursor: sqlite3.Cursor = db_connection.cursor()
        insert_row = (row_to_add.uuid, row_to_add.user_id, row_to_add.title, row_to_add.date, row_to_add.groupID,
                      row_to_add.groupName)

        db_cursor.execute(f"INSERT INTO {self.uploaded_events_table_name} (uuid, user_id, title, date, group_id, group_name)" +
                          f"VALUES (?, ?, ?, ? , ?, ?)", insert_row)
        primary_key = db_cursor.lastrowid
        db_connection.commit()
        return primary_key






import requests

from calendar_event_engine.db.db_cache import SQLiteDB
from calendar_event_engine.logger import create_logger_from_designated_logger
from calendar_event_engine.parser.package import get_group_package
from calendar_event_engine.publishers.abc_publisher import Publisher
from calendar_event_engine.publishers.mobilizon.uploader import MobilizonUploader
from calendar_event_engine.scrapers.abc_scraper import Scraper
from calendar_event_engine.scrapers.google_calendar.scraper import GoogleCalendarScraper
from calendar_event_engine.scrapers.ical.scraper import ICALScraper
from calendar_event_engine.scrapers.statics.scraper import StaticScraper
from calendar_event_engine.types.submission import PublisherTypes, ScraperTypes, GroupPackage
from calendar_event_engine.types.submission_handlers import RunnerSubmission

logger = create_logger_from_designated_logger(__name__)


def get_runner_submission(remote_submission_url: str, test_mode: bool, cache_db: SQLiteDB) -> RunnerSubmission:
    json_submission: dict = requests.get(remote_submission_url).json()

    publisher_package: dict[Publisher, list[GroupPackage]] = dict()
    respective_scrapers: dict[ScraperTypes, Scraper] = dict()
    for publisher in json_submission.keys():
        publisher_instance: Publisher
        match publisher:
            case PublisherTypes.MOBILIZON.value:
                publisher_instance = MobilizonUploader(test_mode, cache_db)
            case _:
                raise TypeError("Expected publisher that is accepted, instead got: " + publisher)
        publisher_package[publisher_instance] = []
        for group_package_source_path in json_submission[publisher]:
            group_package: GroupPackage = get_group_package(group_package_source_path)
            publisher_package[publisher_instance].append(group_package)
            for scraper_type in list(group_package.scraper_type_and_kernels.keys()):
                if scraper_type not in respective_scrapers:
                    logger.info(f"Creating scraper of type: {scraper_type}")
                    match scraper_type:
                        case ScraperTypes.GOOGLE_CAL:
                            respective_scrapers[scraper_type] = GoogleCalendarScraper(cache_db)
                        case ScraperTypes.STATIC:
                            respective_scrapers[scraper_type] = StaticScraper()
                        case ScraperTypes.ICAL:
                            respective_scrapers[scraper_type] = ICALScraper(cache_db)

    return RunnerSubmission(cache_db, publisher_package, test_mode, respective_scrapers)





import json
import os

import validators
from calendar_event_engine.db.db_cache import SQLiteDB
from calendar_event_engine.db.event_source_driver import EventSource
from calendar_event_engine.db.uploaded_events_driver import UploadedEventRow
from calendar_event_engine.logger import create_logger_from_designated_logger
from calendar_event_engine.publishers.abc_publisher import Publisher
from calendar_event_engine.publishers.mobilizon.api import MobilizonAPI
from calendar_event_engine.publishers.mobilizon.types import MobilizonEvent, EventParameters
from calendar_event_engine.types.generics import GenericEvent
from calendar_event_engine.types.submission import AllEventsFromAGroup, GroupEventsKernel

logger = create_logger_from_designated_logger(__name__)

def none_if_not_present(x, dictionary):
    return None if x not in dictionary else dictionary[x]

class MobilizonUploader(Publisher):

    mobilizonAPI: MobilizonAPI
    cache_db: SQLiteDB
    __fakeUUIDForTests = 0

    def __init__(self, test_mode, cache_db):
        super().__init__(cache_db, logger)
        self.testMode = test_mode

    def close(self):
        if not self.testMode:
            self.mobilizonAPI.logout()

    def update(self) -> None:
        pass

    def monitor(self) -> None:
        pass

    def upload_individual_event(self, event_to_upload: GenericEvent) -> dict:
        event: MobilizonEvent = self.generic_event_converter(event_to_upload)
        if self.testMode:
            self.__fakeUUIDForTests += 1
            return {"id": 1, "uuid": self.__fakeUUIDForTests, "groupId": event.attributedToId}
        else:
            upload_response: dict = {}
            if event.picture is not None and validators.url(event.picture.mediaId):
                potential_id = self.mobilizonAPI.upload_file(event.picture.mediaId)
                if potential_id != "":
                    event.picture.mediaId = potential_id
            upload_response = self.mobilizonAPI.bot_created_event(event)
            upload_response["groupId"] = event.attributedToId
            return upload_response

    def create_cachable_response(self, upload_response: dict, event_kernel: GroupEventsKernel, event: GenericEvent, source_id):
        upload_row = UploadedEventRow(uuid=upload_response["uuid"], user_id=upload_response["id"],
                                      title=event.title, date=event.begins_on,
                                      group_id=upload_response["groupId"], group_name=event_kernel.group_name)
        upload_source = EventSource(uuid=upload_response["uuid"], website_url=event.online_address,
                                    calendar_id=source_id, source_type=event_kernel.scraper_type.value)
        return upload_row, upload_source

    def connect(self):
        if not self.testMode:
            endpoint = os.environ.get("MOBILIZON_ENDPOINT")
            email = os.environ.get("MOBILIZON_EMAIL")
            passwd = os.environ.get("MOBILIZON_PASSWORD")

            if email is None and passwd is None:
                login_file_path = os.environ.get("MOBILIZON_LOGIN_FILE")
                with open(login_file_path, "r") as f:
                    secrets = json.load(f)
                    email = secrets["email"]
                    passwd = secrets["password"]

            self.mobilizonAPI = MobilizonAPI(endpoint, email, passwd)

    def generic_event_converter(self, generic_event: GenericEvent):
        mobilizon_metadata = generic_event.publisher_specific_info["mobilizon"]
        category = None if "defaultCategory" not in mobilizon_metadata else EventParameters.Categories[mobilizon_metadata["defaultCategory"]]

        if validators.url(generic_event.picture):
            mobilizon_picture = EventParameters.MediaInput(generic_event.picture)
        else:
            mobilizon_picture = None if "defaultImageID" not in mobilizon_metadata else EventParameters.MediaInput(mobilizon_metadata["defaultImageID"])


        mobilizon_tags = None if "defaultTags" not in mobilizon_metadata else mobilizon_metadata["defaultTags"]
        generic_physical_address = generic_event.physical_address
        mobilizon_physical_address = None if generic_physical_address == None else EventParameters.Address(locality=generic_physical_address.locality,
                                                             postalCode=generic_physical_address.postalCode,
                                                             street=generic_physical_address.street,
                                                             geom=generic_physical_address.geom,
                                                             country=generic_physical_address.country,
                                                             region=generic_physical_address.region,
                                                             timezone=generic_physical_address.timezone,
                                                             description=generic_physical_address.description)
        mobilizon_event = MobilizonEvent(mobilizon_metadata["groupID"], generic_event.title,
                                         description=generic_event.description, beginsOn=generic_event.begins_on,
                                         endsOn=generic_event.ends_on, tags=mobilizon_tags,
                                         onlineAddress=generic_event.online_address, physicalAddress=mobilizon_physical_address,
                                         category=category, picture=mobilizon_picture)
        return mobilizon_event






























import logging


def create_logger_from_designated_logger(logger_name: str, codebase_logger_level: int=logging.INFO):
    logging.basicConfig(format='%(asctime)s - %(levelname)s - %(name)s - %(message)s')
    logger = logging.getLogger("codebase_logger")
    logger.setLevel(codebase_logger_level)
    return logger.getChild(logger_name)





import requests
import validators

from calendar_event_engine.types.generics import GenericEvent
from calendar_event_engine.types.submission import AllEventsFromAGroup


# https://www.geeksforgeeks.org/http-headers-content-type/

def normalize_generic_event(events_to_upload_list: list[AllEventsFromAGroup]):
    allowed_image_content_type_header = ["image/gif", "image/jpeg", "image/png"]
    for events_to_upload in events_to_upload_list:
        generic_event: GenericEvent
        for generic_event in events_to_upload.events:
            # If valid URL but not a picture, set to empty string
            if validators.url(generic_event.picture):
                picture_head = requests.head(generic_event.picture)
                if picture_head.status_code != 200 and picture_head.headers.get("Content-type") not in allowed_image_content_type_header:
                    generic_event.picture = ""

            if not validators.url(generic_event.online_address):
                generic_event.online_address = ""









from enum import Enum

from gql import gql
from pydantic import BaseModel

from calendar_event_engine.publishers.mobilizon.types import MobilizonEvent


# TODO: Clean all text for dangerous input

# EventType = MobilizonTypes.EventType
# https://github.com/framasoft/mobilizon/blob/main/docs/dev.md

# Cleaning Received Text
# https://stackoverflow.com/questions/10993612/how-to-remove-xa0-from-string-in-python

def _conditional_attribute(key: str, value):
  return ((key + ": " + str(value) + ",\n") if value is not None else "")


def conditional_gql_inputs(classDataObject: BaseModel or dict):
  classDict: dict = classDataObject.dict() if isinstance(classDataObject, (BaseModel)) else classDataObject
  gqlString = """"""
  for key, value in classDict.items():
    valueType = type(value)
    if (valueType is str or valueType is int):
      if (valueType is str):
        value = value.replace('"', "'")
        # value = re.sub(r"(\n{2,})|( +(?=\n))", "", value) #No multiple newlines in succession
        value = f'"""{value}"""'
      gqlString += _conditional_attribute(key, value)
    elif (isinstance(value, Enum)):
      value = value.value
      gqlString += _conditional_attribute(key, value)
    elif (valueType is dict and value is not None):
      gqlString = f"""{gqlString}{key}:{{
          {conditional_gql_inputs(value)}
        }},"""
  return gqlString

# conditional_gql_inputs(EventType(title="f", description="f", actorID=1, picture=EventParameters.MediaInput(name="fd", url="bar")))


# ==== GQL : Events ====
class EventGQL:
    def createEventGQL(eventInformation: MobilizonEvent):
        gqlString = f"""
        mutation {{ createEvent(
          {conditional_gql_inputs(eventInformation)}
          )
          {{
            id
            uuid
          }}
        }}
    """
        return gql(gqlString)
    
    def uploadMediaRawGQL():
      gqlString = """
        mutation($file: Upload!, $name: String!, $actorId: ID){
          uploadMedia(
            file: $file,
            name: $name,
            actorId: $actorId
          ){
            id
          }
        }
      """
      return gqlString

#           {("picture: " + eventInformation.picture) + "," if eventInformation.picture != None else ""} 

# ==== GQL : credentials ====


class AuthenticationGQL:
    def loginGQL(email, password):
        gqlString = f"""
      mutation {{
      login(
      email: {email}, 
      password: {password}) {{
        accessToken   
        refreshToken
      }}
    }}
    """
        return gql(gqlString)

    def logoutGQL(refreshToken):
        gqlString = f"""
      mutation {{
      logout(refreshToken: {refreshToken})
    }}
    """
        return gql(gqlString)

    def refreshTokenGQL(refreshToken):
        gqlString = f"""
      mutation {{ RefreshToken(
      refreshToken: {refreshToken}) {{
	    accessToken   
      refreshToken
      }}
    }}"""
        return gql(gqlString)
    

class ActorsGQL:
  def getIdentities():
    gqlString = """
    query {
        identities{
          id
          type
          preferredUsername
          name
          url
        }
      }
    """
    return gql(gqlString)
  
  def getGroups(membershipName, page=1, limit=30):
    gqlString = f"""
    query {{
      loggedUser{{
        id
        memberships(name: {membershipName}, page: {page}, limit: {limit}){{
          total
          elements{{
            role
            actor{{
              id
              type
              preferredUsername
              name
            }}
            parent{{
              id
              type
              preferredUsername
              name
            }}
          }}
        }}
      }}
    }}
    """
    return gql(gqlString)


# UPDATE_GQL = gql("""
# mutation updateEvent($id: ID!, $title: String, $description: String, $beginsOn: DateTime, $endsOn: DateTime, $status: EventStatus, $visibility: EventVisibility, $joinOptions: EventJoinOptions, $draft: Boolean, $tags: [String], $picture: PictureInput, $onlineAddress: String, $phoneAddress: String, $organizerActorId: ID, $attributedToId: ID, $category: String, $physicalAddress: AddressInput, $options: EventOptionsInput, $contacts: [Contact]) {
#   updateEvent(eventId: $id, title: $title, description: $description, beginsOn: $beginsOn, endsOn: $endsOn, status: $status, visibility: $visibility, joinOptions: $joinOptions, draft: $draft, tags: $tags, picture: $picture, onlineAddress: $onlineAddress, phoneAddress: $phoneAddress, organizerActorId: $organizerActorId, attributedToId: $attributedToId, category: $category, physicalAddress: $physicalAddress, options: $options, contacts: $contacts) {
# 	id
# 	uuid
#   }
# }
# """)


# CANCEL_GQL = gql("""
# mutation updateEvent($eventId: ID!) {
#   updateEvent(eventId: $eventId, status: CANCELLED) {
# 	id
# 	uuid
#   }
# }
# """)

# CONFIRM_GQL = gql("""
# mutation updateEvent($eventId: ID!) {
#   updateEvent(eventId: $eventId, status: CONFIRMED) {
# 	id
# 	uuid
#   }
# }
# """)


# DELETE_GQL = gql("""
# mutation DeleteEvent($eventId: ID!) {
#   deleteEvent(eventId: $eventId) {
# 	id
#   }
# }
# """)

# # ==== /GQL : Events ====

# # ==== GQL : Actors ====

# CREATE_USER_GQL = gql("""
# mutation createUser($email: String!, $locale: String="fr", $password: String!) {
#   createUser(email: $email, locale: $locale, password: $password) {
# 	id
#   }
# }""")

# CREATE_GROUP_GQL = gql("""
# mutation createGroup($name: String, $preferredUsername: String!, $summary: String = "") {
#   createGroup(name: $name, preferredUsername: $preferredUsername, summary: $summary) {
# 	id
#   }
# }""")

# CREATE_PERSON_GQL = gql("""
# mutation createPerson($name: String, $preferredUsername: String!, $summary: String = "") {
#   createPerson(name: $name, preferredUsername: $preferredUsername, summary: $summary) {
# 	id   url
#   }
# }""")

# CREATE_MEMBER_GQL = gql("""
# mutation inviteMember($groupId: ID!, $targetActorUsername: String!) {
#   inviteMember(groupId: $groupId, targetActorUsername: $targetActorUsername) {
# 	id   role
#   }
# }""")

# UPDATE_MEMBER_GQL = gql("""
# mutation updateMemberRole($memberId: ID!, $role:MemberRoleEnum!) {
#   updateMember(memberId: $memberId, role: $role) {
# 	id   role
#   }
# }""")

# # ==== /GQL : Users ====

# # ==== /GQL : credentials ====

# # ==== GQL : identities - actors - persons and groups ====

# PROFILES_GQL = gql("""
# query Identities { identities { ...ActorFragment } }
# fragment ActorFragment on Actor { id type preferredUsername name url}
# """)

# GROUPS_GQL = gql("""
# query LoggedUserMemberships($membershipName: String, $page: Int, $limit: Int) {
#   loggedUser {
# 	memberships(name: $membershipName, page: $page, limit: $limit) {
# 	  elements {
# 		role
# 		actor { ...ActorFragment }
# 		parent { ...ActorFragment }
# 	  }
# 	}
#   }
# }
# fragment ActorFragment on Actor {  id type  preferredUsername  name }
# """)


"""
        mutation {{ createEvent(
          organizerActorId: {eventInformation.organizerActorId},
          attributedToId: {eventInformation.attributedToId}, 
          title: {eventInformation.title}, 
          description: {eventInformation.description}, 
          beginsOn: "2020-10-29T00:00:00+01:00"
          endsOn: "2022-03-31T23:59:59+02:00"
          status: {eventInformation.status}, 
          visibility: {eventInformation.visibility}, 
          joinOptions: {eventInformation.joinOptions}, 
          draft: {eventInformation.draft}, 
          {_conditional_attribute("picture", eventInformation.picture)}
          {_conditional_attribute("onlineAddress", eventInformation.onlineAddress)}
          {_conditional_attribute("phoneAddress", eventInformation.phoneAddress)}
          {_conditional_attribute("category", eventInformation.category)}
          physicalAddress: {{
              geom: {eventInformation.physicalAddress.geom},
              locality: {eventInformation.physicalAddress.locality},
              postalCode: {eventInformation.physicalAddress.postalCode},
              street: {eventInformation.physicalAddress.street},
              country: {eventInformation.physicalAddress.country}
            }},
          {_conditional_attribute("contacts", eventInformation.contacts)}
          )
          {{
            id
            uuid
          }}
        }}
    """



import sqlite3
from datetime import datetime
from sqlite3 import Cursor

from calendar_event_engine.db.event_source_driver import EventSourceDriver
from calendar_event_engine.db.uploaded_events_driver import UploadedEventsDriver
from calendar_event_engine.logger import create_logger_from_designated_logger

logger = create_logger_from_designated_logger(__name__)

class CombineDBDriver:
    event_source_table_name: str = EventSourceDriver.event_source_table_name
    uploaded_events_table_name: str = UploadedEventsDriver.uploaded_events_table_name

    allColumns = f"""{uploaded_events_table_name}.uuid, {uploaded_events_table_name}.user_id,
        {uploaded_events_table_name}.title, {uploaded_events_table_name}.date, 
        {uploaded_events_table_name}.group_id, {uploaded_events_table_name}.group_name,
        {event_source_table_name}.websiteURL, {event_source_table_name}.calendar_id, 
        {event_source_table_name}.sourceType"""



    def select_all_rows_with_calendar_id(self, db_connection: sqlite3.Connection, calendar_id: str) -> Cursor:
        db_cursor = db_connection.cursor()
        # Comma at the end of (groupID,) turns it into a tuple
        res = db_cursor.execute(f"""SELECT {self.allColumns} FROM {self.uploaded_events_table_name} 
                                INNER JOIN {self.event_source_table_name} ON 
                                {self.uploaded_events_table_name}.id={self.event_source_table_name}.id
                                WHERE calendar_id = ?""", (calendar_id,))
        return res

    def get_last_event_date_for_source_id(self, db_connection: sqlite3.Connection, calendar_id) -> datetime:
        db_cursor = db_connection.cursor()
        res = db_cursor.execute(f"""SELECT date FROM {self.uploaded_events_table_name}
                                INNER JOIN {self.event_source_table_name} ON 
                                {self.uploaded_events_table_name}.id={self.event_source_table_name}.id
                                WHERE calendar_id = ?
                                ORDER BY date DESC LIMIT 1""", (calendar_id,))
        # Conversion to ISO format does not like the Z, that represents UTC aka no time zone
        # so using +00:00 is an equivalent to it
        date_string = res.fetchone()[0]
        logger.debug(f"Last date found for calendar ID {calendar_id}: {date_string}")
        return datetime.fromisoformat(date_string)

    def entry_already_in_cache(self, db_connection: sqlite3.Connection, date:str, title:str, calendar_id:str) -> bool:
        db_cursor = db_connection.cursor()
        res = db_cursor.execute(f"""SELECT {self.allColumns} FROM {self.uploaded_events_table_name}
                                INNER JOIN {self.event_source_table_name} ON 
                                {self.uploaded_events_table_name}.id={self.event_source_table_name}.id
                                WHERE date = ? AND title = ? AND calendar_id = ?""", (date, title, calendar_id))
        query = res.fetchall()
        if len(query) > 0:
            return True
        return False




from enum import Enum

from pydantic import BaseModel

def _generate_args(localVariables: dict) -> dict:
    args = {}
    for name, value in localVariables.items():
        if (value is not None and name != "self" and name != "__class__"):
            args[name] = value
    return args

class EventParameters:
    class Status(Enum):
        confirmed = "CONFIRMED"
        canceled = "CANCELED"
        tentative = "TENTATIVE"


    class Visibility(Enum):
        public = "PUBLIC"
    
    class JoinOptions(Enum):
        free = "FREE"
        invite = "INVITE"
        restricted = "RESTRICTED"
    
    class Categories(Enum):
        arts = "ARTS".upper()
        book_clubs = "BOOK_CLUBS".upper()
        business = "BUSINESS"
        causes = "CAUSES"
        comedy = "COMEDY"
        crafts = "CRAFTS"
        food_drink = "FOOD_DRINK"
        health = "HEALTH"
        music = "MUSIC"
        community = "COMMUNITY"
        family_education = "FAMILY_EDUCATION"
        fashion_beauty = "FASHION_BEAUTY"
        film_media = "FILM_MEDIA"
        games = "GAMES"
        language_culture = "LANGUAGE_CULTURE"
        learning = "LEARNING"
        lgbtq = "LGBTQ"
        movements_politics = "MOVEMENTS_POLITICS"
        networking = "NETWORKING"
        party = "PARTY"
        performing_visual_arts = "PERFORMING_VISUAL_ARTS"
        pets = "PETS"
        photography = "PHOTOGRAPHY"
        outdoors_adventure = "OUTDOORS_ADVENTURE"
        spirituality_religion_beliefs = "SPIRITUALITY_RELIGION_BELIEFS"
        science_tech = "SCIENCE_TECH"
        sports = "SPORTS"
        theatre = "THEATRE"
        default = "MEETING"

    class Address(BaseModel):
        """Address object that Mobilizon can utilize

        Args:
            geom: Geo-coordinates for the point where the address is
            region: The state of this address
            locality: Town
            postalCode: ZipCode
            street: Street
            country: Country
            originId: The ID affiliated with the api that gave the location info
        """
        geom: str = None
        locality: str  # Town
        postalCode: str # ZipCode
        street: str # Street
        country: str # 
        type: str = None
        region: str = ""
        timezone: str = "America/New_York"
        description: str = ""
        originId: str = None
        
        def __init__(self, locality:str, postalCode:str = "", street:str = "",
                     country:str = "United States", type=None, originId=None,
                     region:str = "Connecticut", geom:str = None,
                     timezone = "America/New_York", description=""):
            args = _generate_args(locals())
            super().__init__(**args)
    
    class MediaInput(BaseModel):
        mediaId: str = None
        
        def __init__(self, mediaId:str):
            args = _generate_args(locals())
            super().__init__(**args)
        


class MobilizonEvent(BaseModel):
    """_summary_

    Args:
        OrganizerActorId: Actor ID
        AttributedToId: Group ID
        Dates: UTC format, that is YYYY-MM-DD HH:MM:SSZ with T in the middle if wanted
    """
    
    organizerActorId: int = 0 # Actor ID
    attributedToId: int = 0 # Group ID
    title: str = ""
    description: str = None
    beginsOn: str = "2020-10-29T00:00:00+01:00"
    endsOn: str = None
    status: EventParameters.Status = EventParameters.Status.confirmed
    visibility: EventParameters.Visibility = EventParameters.Visibility.public
    joinOptions: EventParameters.JoinOptions = EventParameters.JoinOptions.free
    draft: str = None
    tags: list[str] = []
    picture: EventParameters.MediaInput = None # https://github.com/framasoft/mobilizon/blob/main/lib/graphql/schema/media.ex
    onlineAddress: str = None
    phoneAddress: str = None
    category: EventParameters.Categories = None
    physicalAddress: EventParameters.Address = None
    # options: dict = {}
    contacts: str = None
    
    def __init__(self, attributedToId:int, title: str, description: str, 
                          beginsOn:str, onlineAddress:str = None, endsOn:str = None,
                          physicalAddress: EventParameters.Address = None,
                          category: EventParameters.Categories = None, tags:[] = None,
                          picture: EventParameters.MediaInput = None):
        args = _generate_args(locals())
        super().__init__(**args)


class Actor(BaseModel):
    id: int
    name: str
    preferredUsername: str
    type: str
    url: str
    
    def __init__(self, id: int, name: str, preferredUsername: str, type: str, url:str):
        args = _generate_args(locals())
        super().__init__(**args)
    
# print(isinstance(EventParameters.MediaInput(), (BaseModel)))





from geopy import Nominatim
from geopy.exc import GeocoderTimedOut

from calendar_event_engine.logger import create_logger_from_designated_logger
from calendar_event_engine.types.generics import GenericAddress

logger = create_logger_from_designated_logger(__name__)


def _generate_args(local_variables: dict) -> dict:
    args = {}
    for name, value in local_variables.items():
        if value is not None and name != "self" and name != "__class__":
            args[name] = value
    return args

def find_geolocation_from_address(address: GenericAddress,
                                  default_location: GenericAddress,
                                  event_title: str) -> (GenericAddress, str):
    # Address given is default, so don't need to call Nominatim
    default_location_notif = "with unverified location for address. Please check address on their website"
    if default_location == address:
        logger.debug(f"{event_title} location included with calendar, but is same as default location.")
        return default_location, default_location_notif
    try:
        geo_locator = Nominatim(user_agent="Mobilizon Event Bot", timeout=10)
        geo_code_location = geo_locator.geocode(f"{address.street}, {address.locality}, {address.postalCode}")
        if geo_code_location is None:
            return default_location, default_location_notif
        address.geom = f"{geo_code_location.longitude};{geo_code_location.latitude}"
        logger.debug(f"{event_title}: Outsourced location was {address.street}, {address.locality}")
        return address, ""
    except GeocoderTimedOut:
        return default_location, default_location_notif



from abc import ABC, abstractmethod
from logging import Logger

from calendar_event_engine.db.db_cache import SQLiteDB
from calendar_event_engine.db.event_source_driver import EventSource
from calendar_event_engine.db.uploaded_events_driver import UploadedEventRow
from calendar_event_engine.types.generics import GenericEvent
from calendar_event_engine.types.submission import AllEventsFromAGroup, GroupEventsKernel


class Publisher(ABC):
    """
    The resulting website or remote destination where all scraped events will be stored.
    """
    def __init__(self, cache_db: SQLiteDB, logger: Logger):
        self.cache_db = cache_db
        self.logger= logger

    def upload(self, list_of_groups_event: list[AllEventsFromAGroup]) -> list[GenericEvent]:
        """
        Upload events to the publishers, and returns a list of successfully uploaded events.
        :param list_of_groups_event:
        :return:
        """
        events_uploaded = []
        for events_to_upload in list_of_groups_event:
            all_events = events_to_upload.events
            event_kernel = events_to_upload.eventKernel
            source_id = events_to_upload.calendar_id
            for generic_event in all_events:
                try:
                    if not self.cache_db.entry_already_in_cache(generic_event.begins_on, generic_event.title, source_id):
                        upload_response = self.upload_individual_event(generic_event)
                        event_row, event_source = self.create_cachable_response(upload_response, event_kernel, generic_event, source_id)
                        self.cache_db.insert_uploaded_event(event_row, event_source)
                        self.logger.info(f"{generic_event.title}: {upload_response}")
                        events_uploaded.append(generic_event)

                except Exception as e:
                    self.logger.error(f"Unable to upload the following event: {generic_event}", e)

        return events_uploaded

    @abstractmethod
    def create_cachable_response(self, upload_response: dict, event_kernel: GroupEventsKernel,
                                 generic_event: GenericEvent, calendar_id: str) -> (UploadedEventRow, EventSource):
        pass

    @abstractmethod
    def upload_individual_event(self, event_to_upload: GenericEvent) -> dict:
        pass

    @abstractmethod
    def update(self) -> None:
        """
        Update an event that has already been uploaded to the publisher.
        :return:
        """
        pass

    @abstractmethod
    def connect(self) -> None:
        """
        Connect to the external publisher.
        :return:
        """
        pass

    @abstractmethod
    def close(self) -> None:
        """
        Close the connection created with the publisher.
        :return:
        """
        pass

    @abstractmethod
    def monitor(self) -> None:
        """
        Used to check already scraped events if they have changed, and update them if needed.
        """
        pass














import json

import requests
from gql import Client
from gql.transport.aiohttp import AIOHTTPTransport
from gql.transport.requests import RequestsHTTPTransport
from tenacity import retry, retry_if_exception, stop_after_attempt, wait_exponential

import aiohttp
from calendar_event_engine.logger import create_logger_from_designated_logger
from calendar_event_engine.publishers.mobilizon.gql_requests import EventGQL, AuthenticationGQL, ActorsGQL
from calendar_event_engine.publishers.mobilizon.types import MobilizonEvent, Actor

logger = create_logger_from_designated_logger(__name__)


class retry_if_not_exception_type(retry_if_exception):
    """Retries except an exception has been raised of one or more types."""

    def __init__(self, exception_types=Exception):
        self.exception_types = exception_types
        super(retry_if_not_exception_type, self).__init__(
            lambda e: not isinstance(e, exception_types))




class _MobilizonClient:
    class LoginTokens:
        accessToken: str
        refreshToken: str

        def __init__(self, access_token: str, refresh_token: str):
            self.accessToken = access_token
            self.refreshToken = refresh_token

    loginTokens: LoginTokens = None
    client: Client = None
    endpoint: str = None
    headers: dict = None

    def __init__(self, endpoint: str, email: str, password: str):
        self.endpoint = endpoint
        self.client = self._build_client()
        data = self.publish(AuthenticationGQL.loginGQL(email, password))
        login = data['login']
        self.loginTokens = self.LoginTokens(login['accessToken'], login['refreshToken'])
        self.client = self._build_client(self.loginTokens.accessToken)

    def _build_client(self,bearer=None):
        self.headers = dict()
        if bearer is not None:
            self.headers['Authorization'] = 'Bearer ' + bearer

        transport = RequestsHTTPTransport(
            url=self.endpoint,
            headers=self.headers,
            verify=True,
            # retries=3,
        )
        return Client(transport=transport, fetch_schema_from_transport=True)


    def refresh_token(self, refresh_token: str):
        return self.publish(AuthenticationGQL.refreshTokenGQL(refresh_token))

    def log_out(self):
        return self.publish(AuthenticationGQL.logoutGQL(f'"{self.loginTokens.refreshToken}"'))  # void


    # attempts at 0s, 2s, 4s, 8s
    @retry(reraise=True, stop=stop_after_attempt(1),
           wait=wait_exponential(multiplier=2))
    def publish(self, query):
        response = self.client.execute(query)
        return response


class MobilizonAPI:
    _mobilizon_client: _MobilizonClient
    bot_actor: Actor
    
    def __init__(self, endpoint: str, email: str, password: str):
        self._mobilizon_client = _MobilizonClient(endpoint, f'"{email}"', f'"{password}"')
        self.bot_actor = Actor(**self.getActors()["identities"][0])
        logger.info("Logged In Mobilizon")

    
    def bot_created_event(self, event_type: MobilizonEvent) -> dict:
        event_type.organizerActorId = self.bot_actor.id

        event_id = self._mobilizon_client.publish(EventGQL.createEventGQL(event_type))
        return {"id": event_id["createEvent"]["id"], "uuid": event_id["createEvent"]["uuid"]}
    
    def upload_file(self, file_path: str) -> str:
        with requests.get(file_path) as file:
            variables = {"name": "automatically-uploaded-event-bot", "file": "uploaded-file", "actorID": self.bot_actor.id}
            response = requests.post(self._mobilizon_client.endpoint,
                data={
                    "query": EventGQL.uploadMediaRawGQL(),
                    "variables": json.dumps(variables)
                },
                files= {
                    "uploaded-file": file.content
                },
                headers={'Authorization': f'Bearer {self._mobilizon_client.loginTokens.accessToken}',
                         'accept': 'application/json'}
            )
            content = json.loads(response.content)
            if content is None or response.status_code != 200 or "errors" in content:
                return ""
            return content["data"]["uploadMedia"]["id"]

    def logout(self):
        self._mobilizon_client.log_out()
        logger.info("Logged Out")
    
    def getActors(self):
        return self._mobilizon_client.publish(ActorsGQL.getIdentities())
    
    def getGroups(self):
        return self._mobilizon_client.publish(ActorsGQL.getGroups('"eventbot"'))

